#!/usr/bin/env python
#@+leo-ver=4
#@+node:@file gmailfs.py
#@@first
#
#    Copyright (C) 2004  Richard Jones  <richard followed by funny at sign then jones then a dot then name>
#
#    GmailFS - Gmail Filesystem Version 0.2
#
#    This program can be distributed under the terms of the GNU GPL.
#    See the file COPYING.
#

"""
GmailFS provides a filesystem using a Google Gmail account as its storage medium
"""

#@+others
#@+node:imports

from fuse import Fuse
import os
from errno import *
from stat import *

import thread

import libgmail
from constants import *

import sys,traceback,re,string,time,tempfile,array,logging

#@-node:imports

# Setup logging
log = logging.getLogger('gmailfs')
log.setLevel(logging.ERROR)


# Globals
DefaultUsername = 'defaultUser'
DefaultPassword = 'defaultPassword'
DefaultFsname = 'gmailfs'



NumberQueryRetries = 3


def _getMessagesByQuery(ga,queryString):
    tries = 0
    while tries<NumberQueryRetries:
        tries+=1
        try:
            folder = ga.getMessagesByQuery(queryString+" subject:fsname=__startfs__"+fsNameVar+"__endfs__")
            return folder
        except:
            log.error("Exception getting query:"+queryString)
            
NumberSendRetries = 3

#@+node:_sendMessages
def _sendMessage(ga,gmsg):
    """
    Send an already composed GmailComposedMessage message to the users own email address
    """
    
    tries = 0
    while tries<NumberSendRetries:
        tries+=1
        try:
            if ga.sendMessage(gmsg):
                log.debug("Sent message ok")
                return 1
            else:
                log.info("Sent message failed")
        except:
            log.info("Sent message failed")
        
    log.error("Send failed too many times")
    return 0
#@+node:_sendMessages

#@+node:class GmailInode
class GmailInode:

    """
    Class used to store gmailfs inode details
    """

    #@+node:__init__
    def __init__(self, msg,ga):
        try:
            self.ino = 0
            self.mode = 0
            self.dev = 0
            self.nlink = 0
            self.uid = 0
            self.gid = 0
            self.size = 0
            self.atime = 0
            self.mtime = 0
            self.ctime = 0
            self.blocksize = DefaultBlockSize
            self.ga = ga
            self.msg = msg
            m = re.match('path=__startpath__(.*)__endpath__ name=__startname__(.*)__endname__ linkto=__startlink__(.*)__endlink__ inoderef=__(.*)__ fsname=__startfs__(.*)__endfs__',msg.subject)
            matchInode = m.group(4)
            log.debug("trying to get inode with:"+matchInode+" from subject:"+msg.subject)
            self.inode_msg = self.getinodeMsg(matchInode)
            self.setInode()
        except:
            traceback.print_exc(file=sys.stdout)
            log.error("got exception when getmessages")
    #@-node:__init__

    #@+node:getinodeMsg
    def getinodeMsg(self,inodeNumber):
        """
        Get Gmail message handle for the message containing inodeNumber
        """
        
        folder = _getMessagesByQuery(self.ga,'subject:inodeid=__'+inodeNumber+'__')
        if len(folder)!=1:
          log.error("could not find inode msg with inodenumber:"+inodeNumber)
          return None;
        if len(folder._threads[0])!=1:
          return None;
        thread = folder._threads[0]
        if not thread._messages:
          thread._messages = thread._getMessages(thread)

        return thread._messages[0]
    #@-node:getinodeMsg
    
    def update(self):
        """
        Sync's the state of this inode back to the users gmail account
        """
        
        timeString = str(int(time.time()))
        subject = "inodeid=__"+str(self.ino)+"__ mode=__"+str(self.mode)+"__ dev=__"+str(self.dev)+"__ nlink=__"+str(self.nlink)+"__ uid=__"+str(self.uid)+"__ gid=__"+str(self.gid)+"__ size=__"+str(self.size)+"__ atime=__"+str(self.atime)+"__ mtime=__"+timeString+"__ ctime=__"+str(self.ctime)+"__ blocksize=__"+str(self.blocksize)+"__ fsname=__startfs__"+fsNameVar+"__endfs__"
        gmsg = libgmail.GmailComposedMessage(username, subject, "emptymsg")
        if _sendMessage(self.ga,gmsg):
            log.debug("Sent "+subject+" ok")
            self.ga.trashMessage(self.inode_msg)
            self.inode_msg = self.getinodeMsg(str(self.ino))
        else:
            e = OSError("Couldnt send mesg"+path)
            e.errno = ENOSPC
            raise e
        

    def setInode(self):
        """
        Setup the inode instances members from the gmail inode message
        """
        
        try:
            subject = self.inode_msg.subject
            log.debug("setting inode from:"+subject)
            m = re.match('inodeid=__(.*)__ mode=__(.*)__ dev=__(.*)__ nlink=__(.*)__ uid=__(.*)__ gid=__(.*)__ size=__(.*)__ atime=__(.*)__ mtime=__(.*)__ ctime=__(.*)__ blocksize=__(.*)__ fsname=__startfs__(.*)__endfs__', subject)
            self.ino = int(m.group(1))
            self.mode = int(m.group(2))
            self.dev = int(m.group(3))
            self.nlink = int(m.group(4))
            self.uid = int(m.group(5))
            self.gid = int(m.group(6))
            self.size = int(m.group(7))
            self.atime = int(m.group(8))
            self.mtime = int(m.group(9))
            self.ctime = int(m.group(10))
            self.blocksize = int(m.group(11))
        except:
            traceback.print_exc(file=sys.stdout)
            self.ino = None

    def unlink(self, path):
        """
        Delete this inode and all of its data blocks
        """
        
        self.nlink-=1
        self.ga.trashMessage(self.msg)
        if self.nlink < 0:
           self.ga.trashMessage(self.inode_msg)
           subject = 'inodeblockid=__'+str(self.ino)+'__'
           folder = _getMessagesByQuery(self.ga,'subject:'+subject)
           for thread in folder:
               for msg in thread:
                   self.ga.trashMessage(msg)
#@-node:class GmailInode

#@+node:class OpenFile
class OpenFile:
    """
    Class holding any currently open files, includes cached instance of the last data block retrieved
    """
    
    def __init__(self, inode):
        self.inode = inode
        self.tmpfile = None
        self.blocksRead = 0
        self.needsWriting = 0
        self.blocksize = inode.blocksize
        self.buffer = list(" "*self.blocksize)
        self.currentOffset = -1
        self.lastBlock = 0
        self.lastBlockRead = -1
        self.lastBlockBuffer = []

    def close(self):
        """
        Closes this file by committing any changes to the users gmail account
        """
        
        if self.needsWriting:
           self.commitToGmail()


    def write(self,buf,off):
        """
        Write data to file from buf, offset by off bytes into the file
        """
        
        buflen = len(buf)
        towrite = buflen
        #if self.currentOffset == -1 or off<self.currentOffset or off>self.currentOffset+self.blocksize:
        #    self.commitToGmail()
        #    self.currentOffset = (off/self.blocksize)*self.blocksize+(off/self.blocksize)
        #    self.buffer = self.readFromGmail(self.currentOffset/self.blocksize,1)

        if self.currentOffset == -1 or off<self.currentOffset or off>self.currentOffset:
            self.commitToGmail()
            self.currentOffset = off;
            self.buffer = self.readFromGmail(self.currentOffset/self.blocksize,1)
            
        currentBlock = self.currentOffset/self.blocksize
        written = 0
        while towrite>0:
            thiswrote = min(towrite,min(self.blocksize-(self.currentOffset%self.blocksize),self.blocksize))
            log.debug("wrote "+str(thiswrote)+" bytes off:"+str(off)+" self.currentOffset:"+str(self.currentOffset))
            self.buffer[self.currentOffset%self.blocksize:] = buf[written:written+thiswrote]
            towrite -= thiswrote
            written += thiswrote
            self.currentOffset += thiswrote
            self.lastBlock=currentBlock
            if self.currentOffset/self.blocksize>currentBlock:
                self.needsWriting = 1
                self.commitToGmail()
                currentBlock+=1
                if towrite>0:
                    self.buffer = self.readFromGmail(currentBlock,1)
            else:
                self.needsWriting = 1

        if off+buflen>self.inode.size:
            self.inode.size=off+buflen
            
        return buflen

    def commitToGmail(self):
        """
        Send any unsaved data to users gmail account as an attachment
        """
        
        if not self.needsWriting:
            return 1
        
        ga = self.inode.ga
        subject = 'inodeblockid=__'+str(self.inode.ino)+'__ blocknumber=__'+str(self.lastBlock)+'__ fsname=__startfs__'+fsNameVar+'__endfs__'
        tmpf = tempfile.NamedTemporaryFile()

        arr = array.array('c')
        arr.fromlist(self.buffer)
        if log.isEnabledFor(logging.DEBUG):
            log.debug("wrote to tmp file:"+arr.tostring())
        tmpf.write(arr.tostring())
        tmpf.flush()
        gmsg = libgmail.GmailComposedMessage(username, subject, "emptymsg",filenames = [tmpf.name])
        if _sendMessage(ga,gmsg):
            log.debug("Sent write commit ok")
            self.needsWriting=0
            self.inode.update()
            tmpf.close()
            return 1
        else:
            log.error("Sent write commit failed")
            tmpf.close()
            return 0

    

    def read(self,readlen,offset):
        """
        Read readlen bytes from an open file from position offset bytes into the files data
        """
        
        readlen = min(self.inode.size-offset,readlen)
        outbuf = list(" "*readlen)
        toread = readlen;
        upto=0;
        while toread>0:
           readoffset = (offset+upto)%self.blocksize
           thisread = min(toread,min(self.blocksize-(readoffset%self.blocksize),self.blocksize))
           outbuf[upto:] = self.readFromGmail((offset+upto)/self.blocksize,0)[readoffset:readoffset+thisread]
           upto+=thisread
           toread-=thisread
           log.debug("still to read:"+str(toread))
        return outbuf

        return retbuf

    def readFromGmail(self,readblock,deleteAfter):
        """
        Read data block with block number 'readblock' for this file from users gmail account, if 'deleteAfter' is
        true then the block will be removed from Gmail after reading
        """
        
        log.debug("about to try and find inode:"+str(self.inode.ino)+" blocknumber:"+str(readblock))
        if self.lastBlockRead == readblock:
            return self.lastBlockBuffer
        ga = self.inode.ga
        subject = 'inodeblockid=__'+str(self.inode.ino)+'__ blocknumber=__'+str(readblock)+'__'
        folder = _getMessagesByQuery(ga,'subject:'+subject)
        if len(folder)!=1:
            return list(" "*self.blocksize)
        if len(folder._threads[0])!=1:
           return list(" "*self.blocksize)
        thread = folder._threads[0]
        if not thread._messages:
            thread._messages = thread._getMessages(thread)
        msg = thread._messages[0]
        log.debug("got msg with subject:"+msg.subject)
        log.debug("got "+str(len(msg.attachments))+" attachments")
        a = msg.attachments[0]
        if log.isEnabledFor(logging.DEBUG):
            log.debug("read from gmail:"+a.content)
        if deleteAfter:
            ga.trashMessage(msg)
        self.lastBlockRead = readblock
        self.lastBlockBuffer = a.content
        contentList = list(" "*self.blocksize)
        contentList[0:] = a.content

        return contentList

#@-node:class OpenFile

#@+node:class Gmailfs
class Gmailfs(Fuse):

    #@	@+others

    
    #@+node:__init__
    def __init__(self, *args, **kw):
    
        Fuse.__init__(self, *args, **kw)
    
        if 0:
            print "gmailfs.py:Gmailfs:mountpoint: %s" % repr(self.mountpoint)
            print "gmailfs.py:Gmailfs:unnamed mount options: %s" % self.optlist
            print "gmailfs.py:Gmailfs:named mount options: %s" % self.optdict
    
        # do stuff to set up your filesystem here, if you want

        self.openfiles = {}
        self.inodeCache = {}

        global DefaultBlockSize
        
        global fsNameVar
        global password
        global username

        DefaultBlockSize = 5*1024*1024
        
        fsNameVar = DefaultFsname
        password = DefaultPassword
        username = DefaultUsername

        if not self.optdict.has_key("username"):
            print 'mount: warning, should mount with username=gmailuser option, using default'
        else:
            username = self.optdict['username']

        if not self.optdict.has_key("password"):
            print 'mount: warning, should mount with password=gmailpass option, using default'
        else:
            password = self.optdict['password']

        if not self.optdict.has_key("fsname"):
            print 'mount: warning, should mount with fsname=name option, using default'
        else:
            fsNameVar = self.optdict['fsname']

        if self.optdict.has_key("blocksize"):
            DefaultBlockSize = int(self.optdict['blocksize'])
        
        self.ga = libgmail.GmailAccount(username, password)
        self.ga.login()
        log.info("Connected to gmail")
        #thread.start_new_thread(self.mythread, ())
        pass
    #@-node:__init__

    #@+node:mythread
    def mythread(self):
    
        """
        The beauty of the FUSE python implementation is that with the python interp
        running in foreground, you can have threads
        """    
        log.debug("mythread: started")
        #while 1:
        #    time.sleep(120)
        #    print "mythread: ticking"
    
    #@-node:mythread

    #@+node:attribs
    flags = 1
    
    #@-node:attribs
    
    #@+node:getattr
    def getattr(self, path):
        log.debug("get stat:"+path)
        #st_mode (protection bits)
        #st_ino (inode number)
        #st_dev (device)
        #st_nlink (number of hard links)
        #st_uid (user ID of owner)
        #st_gid (group ID of owner)
        #st_size (size of file, in bytes)
        #st_atime (time of most recent access)
        #st_mtime (time of most recent content modification)
        #st_ctime (time of most recent content modification or metadata change).
        #print os.lstat(path)

        if path == '/':
            return (S_IFDIR|S_IRUSR|S_IXUSR|S_IWUSR|S_IRGRP|S_IXGRP|S_IXOTH|S_IROTH,11,0,0,0,0,5555,1000,1000000,10000000)
        inode = self.getinode(path)
        if inode:
            if log.isEnabledFor(logging.DEBUG):
                log.debug("inode "+str(inode))
            statTuple = (inode.mode,inode.ino,inode.dev,inode.nlink,inode.uid,inode.gid,inode.size,inode.atime,inode.mtime,inode.ctime)
            if log.isEnabledFor(logging.DEBUG):
                log.debug("statsTuple "+str(statTuple))
            return statTuple
        else:
            e = OSError("No such file"+path)
            e.errno = ENOENT
            raise e
    #@-node:getattr
    
    #@+node:readlink
    def readlink(self, path):
        log.debug("readlink: path='%s'" % path)
        inode = self.getinode(path)
        if not (inode.mode & S_IFLNK):
            e = OSError("Not a link"+path)
            e.errno = EINVAL
            raise e
        log.debug("about to follow link in subject:"+inode.msg.subject)
        m = re.search('linkto=__startlink__(.*)__endlink__',inode.msg.subject)
        return m.group(1)
    #@-node:readlink
    
    #@+node:getdir
    def getdir(self, path):
        try:
            log.debug("getting dir "+path)
            fspath = re.sub("/","__fs__",path)
            log.debug("querying for:"+'subject:path=__startpath__'+fspath+'__endpath__')
            # FIX need to check if directory exists and return error if it doesnt, actually
            # this may be done for us by fuse
            folder = _getMessagesByQuery(self.ga,'subject:path=__startpath__'+fspath+'__endpath__')
            log.debug("got folder ")
            lst = []
            for thread in folder:
               for msg in thread:
                 m = re.match('path=__startpath__(.*)__endpath__ name=__startname__(.*)__endname__',msg.subject)
                 matchPath = re.sub("__fs__","/",m.group(1))
                 if log.isEnabledFor(logging.DEBUG):
                     log.debug("Found file:"+matchPath+m.group(2))
                 lst.append(m.group(2))
        except:
            traceback.print_exc(file=sys.stdout)
            log.error("got exception when getmessages")
            lst = None
        return map(lambda x: (x,0), lst)
    #@-node:getdir
    
    #@+node:unlink
    def unlink(self, path):
        log.debug("unlink called on:"+path)
        try:
            inode = self.getinode(path)
            inode.unlink(path)
            del self.inodeCache[path]
            return 0
        except:
            traceback.print_exc(file=sys.stdout)
            e = OSError("Error unlinking file"+path)
            e.errno = EINVAL
            raise e
    #@-node:unlink

    #@+node:rmdir
    def rmdir(self, path):
        log.debug("rmdir called on:"+path)
        #this is already checked before rmdir is even called
        #dirlist = self.getdir(path)
        #if len(dirlist)>0:
        #    e = OSError("directory not empty"+path)
        #    e.errno = ENOTEMPTY
        #    raise e
        inode = self.getinode(path)
        inode.unlink(path)
        del self.inodeCache[path]
        return 0
        
    #@-node:rmdir
    
    #@+node:symlink
    def symlink(self, path, path1):
        log.debug("symlink: path='%s', path1='%s'" % (path, path1))
        self._mkfileOrDir(path1,path,S_IFLNK|S_IRWXU|S_IRWXG|S_IRWXO,-1)
    #@-node:symlink

    #@+node:rename
    def rename(self, path, path1):
        log.debug("rename from:"+path+" to:"+path1)
        msg = self.getinodemsg(path)

        ind = string.rindex(path1,'/')
        log.debug("ind:"+str(ind))
        dirpath = path1[:ind]
        if len(dirpath)==0:
            dirpath="/"
        name = path1[ind+1:]
        log.debug("dirpath:"+dirpath+" name:"+name)
        fspath = re.sub("/","__fs__",dirpath)
        
        m = re.match('path=__startpath__(.*)__endpath__ name=__startname__(.*)__endname__ linkto=__startlink__(.*)__endlink__ inoderef=__(.*)__ fsname=__startfs__(.*)__endfs__',msg.subject)
        subject = "path=__startpath__"+fspath+"__endpath__ name=__startname__"+name+"__endname__ linkto=__startlink__"+m.group(3)+"__endlink__ inoderef=__"+m.group(4)+"__ fsname=__startfs__"+fsNameVar+"__endfs__"
        gmsg = libgmail.GmailComposedMessage(username, subject, "emptymsg")
        if _sendMessage(self.ga,gmsg):
            log.debug("Sent "+subject+" ok")
            self.ga.trashMessage(msg)
            if self.inodeCache.has_key(path):
                del self.inodeCache[path]
            return 0
        else:
            e = OSError("Couldnt send mesg"+path)
            e.errno = ENOSPC
            raise e

    #@-node:rename
    
    #@+node:link
    def link(self, path, path1):
        log.debug("hard link: path='%s', path1='%s'" % (path, path1))
        inode = self.getinode(path)
        if not (inode.mode & S_IFREG):
            e = OSError("hard links only supported for regular files not directories:"+path)
            e.errno = EPERM
            raise e
        inode.nlink+=1
        inode.update()
        self._mkfileOrDir(path1,None,inode.mode,inode.ino)
        return 0
    #@-node:link
    
    #@+node:chmod
    def chmod(self, path, mode):
        log.debug("chmod called with path:"+path+" mode:"+str(mode))
        inode = self.getinode(path)
        inode.mode = (inode.mode & ~(S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)) | mode
        inode.update()
        return 0
    #@-node:chmod

    #@+node:chown
    def chown(self, path, user, group):
        log.debug("chown called with user:"+str(user)+" and group:"+str(group))
        inode = self.getinode(path)
        inode.uid = user
        inode.gid = group
        inode.update()
        return 0
    #@-node:chown

    #@+node:truncate
    def truncate(self, path, size):
        log.debug("truncate "+path+" to size:"+str(size))
        inode = self.getinode(path)
        # this is VERY lazy, we leave the truncated data around
        # it WILL be harvested when we grow the file again or
        # when we delete the file but should probably FIX
        inode.size = size;
        inode.update()
        return 0
    #@-node:truncate

    #@+node:mknod
    def mknod(self, path, mode, dev):
    	""" Python has no os.mknod, so we can only do some things """
        log.debug("mknod:"+path)
    	if S_ISREG(mode):
            self._mkfileOrDir(path,None,mode,-1)
    	    #open(path, "w")
    	else:
    		return -EINVAL
    #@-node:mknod

    def _mkfileOrDir(self,path,path2,mode,inodenumber):
        ind = string.rindex(path,'/')
        log.debug("ind:"+str(ind))
        dirpath = path[:ind]
        if len(dirpath)==0:
            dirpath="/"
        name = path[ind+1:]
        log.debug("dirpath:"+dirpath+" name:"+name)
        fspath = re.sub("/","__fs__",dirpath)
        if path2 == None:
            path2 = ""
        if inodenumber==-1:
            inodeno = int(time.time())
        else:
            inodeno = inodenumber
            
        subject = "path=__startpath__"+fspath+"__endpath__ name=__startname__"+name+"__endname__ linkto=__startlink__"+path2+"__endlink__ inoderef=__"+str(inodeno)+"__ fsname=__startfs__"+fsNameVar+"__endfs__"
        gmsg = libgmail.GmailComposedMessage(username, subject, "emptymsg")
        if _sendMessage(self.ga,gmsg):
            log.debug("Sent "+subject+" ok")
        else:
            e = OSError("Couldnt send mesg"+path)
            e.errno = ENOSPC
            raise e
        # only create inode if number not provided
        if inodenumber==-1:
            timeString = str(int(time.time()))
            subject = "inodeid=__"+str(inodeno)+"__ mode=__"+str(mode)+"__ dev=__11__ nlink=__0__ uid=__0__ gid=__0__ size=__0__ atime=__"+timeString+"__ mtime=__"+timeString+"__ ctime=__"+timeString+"__ blocksize=__"+str(DefaultBlockSize)+"__ fsname=__startfs__"+fsNameVar+"__endfs__"
            gmsg = libgmail.GmailComposedMessage(username, subject, "emptymsg")
            if _sendMessage(self.ga,gmsg):
                log.debug("Sent "+subject+" ok")
            else:
                e = OSError("Couldnt send mesg"+path)
                e.errno = ENOSPC
                raise e

    #@+node:mkdir
    def mkdir(self, path, mode):
        log.debug("mkdir path:"+path+" mode:"+str(mode))
        self._mkfileOrDir(path,None,mode|S_IFDIR, -1)
    #@-node:mkdir

    #@+node:utime
    def utime(self, path, times):
        log.debug("utime for path:"+path+" times:"+str(times))
        inode = self.getinode(path)
        inode.atime = times[0]
        inode.mtime = times[1]
        return 0
    #@-node:utime

    #@+node:open
    def open(self, path, flags):
        log.debug("gmailfs.py:Gmailfs:open: %s" % path)
        try:
            inode = self.getinode(path)
            f = OpenFile(inode)
            self.openfiles[path] = f
            return 0
        except:
            traceback.print_exc(file=sys.stdout)
            e = OSError("Error opening file"+path)
            e.errno = EINVAL
            raise e
    #@-node:open

    #@+node:read
    def read(self, path, readlen, offset):
        try:
            log.debug("gmailfs.py:Gmailfs:read: %s" % path)
            log.debug("reading len:"+str(readlen)+" offset:"+str(offset))
            f = self.openfiles[path]
            buf = f.read(readlen,offset)
            arr = array.array('c')
            arr.fromlist(buf)
            rets = arr.tostring()

            return rets
        except:
            traceback.print_exc(file=sys.stdout)
            e = OSError("Error reading file"+path)
            e.errno = EINVAL
            raise e
    
    #@-node:read

    #@+node:write
    def write(self, path, buf, off):
        try:
            log.debug("gmailfs.py:Gmailfs:write: %s" % path)
            if log.isEnabledFor(logging.DEBUG):
                log.debug("writing:"+str(buf))
            f = self.openfiles[path]
            written = f.write(buf,off)
    	    return written
        except:
            traceback.print_exc(file=sys.stdout)
            e = OSError("Error opening file"+path)
            e.errno = EINVAL
            raise e
    #@-node:write

    #@+node:release
    def release(self, path, flags):
        log.debug("gmailfs.py:Gmailfs:release: %s %s" % (path, flags))
        f = self.openfiles[path]
        f.close()
        del self.openfiles[path]
        return 0
    #@-node:release

    #@+node:statfs
    def statfs(self):
        """
        Should return a tuple with the following 6 elements:
            - blocksize - size of file blocks, in bytes
            - totalblocks - total number of blocks in the filesystem
            - freeblocks - number of free blocks
            - totalfiles - total number of file inodes
            - freefiles - nunber of free file inodes
    
        Feel free to set any of the above values to 0, which tells
        the kernel that the info is not available.
        """
       
        blocks_size = 1024
        blocks = 1000*1024
        quotaInfo = self.ga.getQuotaInfo()
        quotaMbUsed = quotaInfo[QU_SPACEUSED]
        quotaMbTotal = quotaInfo[QU_QUOTA]
        quotaPercent = quotaInfo[QU_PERCENT]
        log.debug("%s of %s used. (%s)\n" % (quotaMbUsed, quotaMbTotal, quotaPercent))
        blocks_free = blocks-(int(quotaMbUsed[:-3])*1024)
        files = 0
        files_free = 0
        namelen = 80
        return (blocks_size, blocks, blocks_free, files, files_free, namelen)
    #@-node:statfs

    #@+node:fsync
    def fsync(self, path, isfsyncfile):
        log.debug("gmailfs.py:Gmailfs:fsync: path=%s, isfsyncfile=%s" % (path, isfsyncfile))
        inode = self.getinode(path)
        f = self.openfiles[path]
        f.commitToGmail()
        return 0
    
    #@-node:fsync


    def getinodemsg(self, path):
      try:
          log.debug("check getnodemsg:"+path)
          ind = string.rindex(path,'/')
          log.debug("ind:"+str(ind))
          dirpath = path[:ind]
          if len(dirpath)==0:
              dirpath="/"
          name = path[ind+1:]
          log.debug("dirpath:"+dirpath+" name:"+name)
          fspath = re.sub("/","__fs__",dirpath)
          folder = _getMessagesByQuery(self.ga,'subject:path=__startpath__'+fspath+'__endpath__ subject:name=__startname__'+name+'__endname__')
          if len(folder)!=1:
              return None;
          if len(folder._threads[0])!=1:
              return None;
          thread = folder._threads[0]
          if not thread._messages:
            thread._messages = thread._getMessages(thread)
          return thread._messages[0]
      except:
          traceback.print_exc(file=sys.stdout)
          log.error("no slash in path:"+path)
          return None


    def getinode(self, path):
        if self.inodeCache.has_key(path):
            return self.inodeCache[path]
        
        msg = self.getinodemsg(path)

        if msg == None:
            return None

        inode = GmailInode(msg,self.ga)
        if inode:
            self.inodeCache[path] = inode
            
        return inode
    
    #@-others
    
#@-node:class Gmailfs
#@+node:mainline
import profile

if __name__ == '__main__':

	server = Gmailfs()
	server.flags = 0
	server.multithreaded = 1;
	server.main()
#@-node:mainline
#@-others
#@-node:@file gmailfs.py
#@-leo
